/* 
 * rlocate.c
 * kernel module for 2.4.x kernels
 *
 * Copyright (C) 2004 Rasto Levrinc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
        #include <linux/modversions.h>
        #define MODVERSIONS
#endif
#include <linux/kernel.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <sys/syscall.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <asm/uaccess.h>

#include "rlocate.h"

MODULE_AUTHOR("Rasto Levrinc");
MODULE_DESCRIPTION("rlocate");
MODULE_LICENSE("GPL");

/* device file */
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);

static int Major;           /* holds major number from device */
static int Device_Open = 0; /* 1 if module is loaded */ 
static char *tmp_path;      /* is used for path */

static struct file_operations fops = {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,10)
        .owner = THIS_MODULE,
#endif
        .read           = device_read,
        .open           = device_open,
        .release        = device_release,
};

extern void* sys_call_table[];
/* overwriten system calls. This wouldn't work for 2.6 kernels. */
asmlinkage long (*orig_link) (const char *oldpath, const char *newpath);
asmlinkage long (*orig_mkdir) (const char *filename, mode_t mode);
asmlinkage long (*orig_mknod) (const char *filename, mode_t mode, dev_t dev);
asmlinkage long (*orig_open) (const char *filename, int flags, mode_t mode);
asmlinkage long (*orig_rename) (const char *oldpath, const char *newpath);
asmlinkage long (*orig_rmdir) (const char *path);
asmlinkage long (*orig_symlink) (const char *oldpath, const char *newpath);
asmlinkage long (*orig_unlink) (const char *pathname);

/* --------------------------- Functions --------------------------- */

/* filenames list functions */



/*
 * filenames_entry_a() puts file name, that was added to the file system, 
 * to the filenames list.
 * It also adds newline to the end and 'a' to the beginning of the path.
 * 'a' stands for 'add'. The first slash is always there so it is
 * redundant 
 * and is removed.
 */
void filenames_entry_a(const char *filename) 
{
        struct nameidata nd;
        int upw_error = user_path_walk_link(filename, &nd);
        if (upw_error) {
                return;
        }

        char *path = d_path(nd.dentry, nd.mnt, tmp_path, PAGE_SIZE);
        if (IS_ERR(path)) {
                return;
        }
        if (strncmp(path, "/tmp/",5) !=0 ) { // ignore /tmp/ directory
                struct filenames *f = kmalloc(sizeof(f), GFP_ATOMIC);
                if (f) {/* don't add to the list if there are allocation error*/
                        char *msg = kmalloc(strlen(path)+2, GFP_ATOMIC);
                        if (msg) { 
                                sprintf(msg, "%s\n", path);
                                *msg = 'a';
                                f->msg = msg;
                                list_add_tail(&f->list, &filenames_list);
                        } else {
                                printk("rlocate: filenames allocation error\n");
                                kfree(f);
                        }
                } else {
                        printk("rlocate: filenames allocation error\n");
                }
        }
        path_release(&nd);
}

/*
 * filenames_entry_r() puts path, that was removed from the file
 * system, to the filenames list. Other than filenames_entry_a, the whole path
 * is an argument to this function.
 * It also adds newline to the end and 'r' to the beginning of the path.
 * 'r' stands for 'remove'. The first slash is always there so it is redundant 
 * and is removed.
 */
void filenames_entry_r(const char *path) 
{
        if (strncmp(path, "/tmp/",5) !=0 ) { // ignore /tmp/ directory
                struct filenames *f = kmalloc(sizeof(f), GFP_ATOMIC);
                if (f) { /* don't add to the list if there is allocation error*/
                        char *msg = kmalloc(strlen(path)+2, GFP_ATOMIC);
                        if (msg) {
                                sprintf(msg, "%s\n", path);
                                *msg = 'r';
                                f->msg = msg;
                                list_add_tail(&f->list, &filenames_list);
                        } else {
                                kfree(f);
                        }
                }
        }
}

/*
 * remove_filenames_entry() removes the filenames_entry from the filenames
 * list and frees the memory
 */
inline void remove_filenames_entry(struct filenames *filenames_entry)
{
        list_del(&filenames_entry->list);
        kfree(filenames_entry->msg);
        kfree(filenames_entry);
}

/*
 * remove_all_filenames_entries() loops over filenames list and removes them
 * all.
 */
inline void remove_all_filenames_entries()
{
        while (!list_empty(&filenames_list)) {
                struct filenames *filenames_entry;
                filenames_entry= list_entry(filenames_list.next,
                                           struct filenames,
                                           list);
                remove_filenames_entry(filenames_entry);
        }
}

/* device file functions */

/*
 * device_open()
 */
static int device_open(struct inode *inode, struct file *file)
{
        if (Device_Open) return -EBUSY;
        Device_Open++;
        MOD_INC_USE_COUNT;

        return SUCCESS;
}

/*
 * device_release()
 */
static int device_release(struct inode *inode, struct file *file)
{
        Device_Open--;
        MOD_DEC_USE_COUNT;
        return 0;
}

/* 
 * device_read() 
 * traverse the list of filenames and write them to the user buffer.
 * Remove every entry that has been written. 
 */
static ssize_t device_read(struct file *file, 
                           char *buffer, 
                           size_t length, 
                           loff_t *offset)
        {
	ssize_t bytes_read = 0;
        char *msg_ptr;
        int ok = 1; // is set to 0 if the line will be outside of the buffer
        while ( (!list_empty(&filenames_list)) && (ok>0) ) {
                struct filenames *f_entry = list_entry(filenames_list.next,
                                                       struct filenames,
                                                       list);
                msg_ptr = f_entry->msg;
                if ( length - bytes_read <= strlen(msg_ptr) ) {
                        ok = 0;
                } else {
	                while (*msg_ptr) {
		                put_user(*(msg_ptr++), buffer++);
		                bytes_read++;
	                }
                        remove_filenames_entry(f_entry);
                }
        }
        return bytes_read;
}

asmlinkage long rlocate_sys_open(const char * filename, 
			   int flags, 
			   int mode)
{
  	struct nameidata nd;
  	int upw_error = user_path_walk_link(filename, &nd);
        
        if (!upw_error) 
                path_release(&nd);
	
        long ret = (*orig_open) (filename, flags, mode);

        if ( upw_error && (ret >= 0) && (flags & O_CREAT) ) { 
                // file didn't exist, there was no error and file was created
	        filenames_entry_a(filename);
        }
  	return (ret);
}

long rlocate_sys_mkdir (const char *filename, mode_t mode)
{
  	long ret = (*orig_mkdir) (filename, mode);

  	if (ret == 0) {
  		filenames_entry_a(filename);
  	}
  	return ret;
}

asmlinkage long rlocate_sys_mknod (const char *filename, mode_t mode, dev_t dev)
{
        long ret = (*orig_mknod) (filename, mode, dev);

        if (ret == 0) 
  		filenames_entry_a(filename);
        
        return ret;
}

asmlinkage long rlocate_sys_rmdir (const char *filename)
{

        struct nameidata nd;
        int upw_error = user_path_walk_link(filename, &nd);
        char *path = d_path(nd.dentry, nd.mnt, tmp_path, PAGE_SIZE);

  	long ret = (*orig_rmdir) (filename);

  	if ( (ret == 0) && (!IS_ERR(path)) ) 
  		filenames_entry_r(path);
  	
        if (!upw_error) 
                path_release(&nd);
        
  	return ret;
}

asmlinkage long rlocate_sys_rename (const char *oldfilename, 
                                    const char *newfilename)
{

        struct nameidata nd;
        int upw_error = user_path_walk_link(oldfilename, &nd);
        char *oldpath = d_path(nd.dentry, nd.mnt, tmp_path, PAGE_SIZE);

  	long ret = (*orig_rename) (oldfilename, newfilename);

  	if (ret == 0) {
                if (!IS_ERR(oldpath))
  		        filenames_entry_r(oldpath);
		filenames_entry_a(newfilename);
	}
        if (!upw_error) 
                path_release(&nd);
  	return ret;
}

asmlinkage long rlocate_sys_symlink (const char *oldpath, const char *newpath)
{
	long ret = (*orig_symlink) (oldpath, newpath);
  	if (ret == 0) 
  		filenames_entry_a(newpath);
  	return ret;
}

asmlinkage int rlocate_sys_link (const char *oldpath, const char *newpath)
{
	long ret = (*orig_link) (oldpath, newpath);
  	if (ret == 0) 
  		filenames_entry_a(newpath);
  	return ret;
}

asmlinkage long rlocate_sys_unlink (const char *filename)
{

        struct nameidata nd;
        int upw_error = user_path_walk_link(filename, &nd);
        char *path = d_path(nd.dentry, nd.mnt, tmp_path, PAGE_SIZE);

	long ret = (*orig_unlink) (filename);

  	if ( (ret == 0) && (!IS_ERR(path)) ) 
  		filenames_entry_r(path);
        if (!upw_error) 
                path_release(&nd);
  	return ret;
}


static int __init init_rlocate(void)
{
	printk(KERN_INFO "rlocate: driver loaded\n");
        Major = register_chrdev(0, DEVICE_NAME, &fops);
        if (Major < 0) {
                printk (KERN_ERR "Registering rlocate failed with %d\n", Major);
                return Major;
        }

        printk (KERN_INFO "rlocate: registered device " 
                DEVICE_NAME " major: %d\n", Major);

        tmp_path = (char*)__get_free_page(GFP_KERNEL); 
        if (!tmp_path) {
                printk (KERN_ERR "rlocate: __get_free_page failed\n");
                return -ENOMEM;
        }

  	orig_link                       = sys_call_table[SYS_link];
  	orig_mkdir                      = sys_call_table[SYS_mkdir];
  	orig_mknod                      = sys_call_table[SYS_mknod];
  	orig_open                       = sys_call_table[SYS_open];
  	orig_rmdir                      = sys_call_table[SYS_rmdir];
  	orig_rename                     = sys_call_table[SYS_rename];
  	orig_symlink                    = sys_call_table[SYS_symlink];
  	orig_unlink                     = sys_call_table[SYS_unlink];

  	sys_call_table[SYS_mkdir]       = rlocate_sys_mkdir;
  	sys_call_table[SYS_mknod]       = rlocate_sys_mknod;

  	sys_call_table[SYS_open]        = rlocate_sys_open; 
  	sys_call_table[SYS_rmdir]       = rlocate_sys_rmdir;
  	sys_call_table[SYS_rename]      = rlocate_sys_rename;
  	sys_call_table[SYS_symlink]     = rlocate_sys_symlink;
  	sys_call_table[SYS_link]        = rlocate_sys_link;
  	sys_call_table[SYS_unlink]      = rlocate_sys_unlink; 

  	return 0;
}

static void __exit exit_rlocate(void)
{

  	printk(KERN_INFO "rlocate: unloading\n");
        remove_all_filenames_entries();
        free_page((unsigned long)tmp_path); 

        int ret = unregister_chrdev(Major, DEVICE_NAME);
        if (ret < 0) printk(KERN_ERR "rlocate: unregister device " 
                            DEVICE_NAME "failed: %d\n", ret);

  	sys_call_table[SYS_link]        = orig_link;
  	sys_call_table[SYS_mkdir]       = orig_mkdir;
  	sys_call_table[SYS_mknod]       = orig_mknod;
  	sys_call_table[SYS_open]        = orig_open;
  	sys_call_table[SYS_rmdir]       = orig_rmdir;
  	sys_call_table[SYS_rename]      = orig_rename;
  	sys_call_table[SYS_symlink]     = orig_symlink;
  	sys_call_table[SYS_unlink]      = orig_unlink;
}


module_init(init_rlocate);
module_exit(exit_rlocate);
